/* burgers.c - some finite differences discretisations of Burgers' SPDE
 *
 * Copyright (C) 2009  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#include <stdio.h>
#include <assert.h>

#include "burgers.h"


int
main()
{
  double *x, *y;
  const double  gamma = THETA*DT*PARAM_nu/(DU*DU);
  const double  delta = (1-THETA)*DT*PARAM_nu/(DU*DU);
  const double  spde_sigma = PARAM_sigma * sqrt(DT/DU);
  unsigned long int  k, kk;
  int  i;

  assert(5*DT < DU);	    /* plausibility check for CFL condition */

  /* initialisation */
  jvrand_init(SEED);
  x = xnew(double, NU);
  y = xnew(double, NU);
  for (i=0; i<NU; ++i)  x[i] = 0;

  /* main loop */
  kk = TMAX/DT+0.5;
  for (k=0; k<kk; ++k) {
    double  xl, xm, xr;

    /* apply (I + nu (1-theta) DT L) */
    xmul(NU, delta, 1-2*delta, delta, x, y);

    /* add drift and noise */
    xm = x[NU-1], xr = x[0];
    for (i=0; i<NU; ++i) {
      xl = xm, xm = xr, xr = (i+1<NU) ? x[i+1] : x[0];
#if METHOD == 1
      y[i] += -xm*(xr-xm)*(DT/DU); /* right hand derivative */
#elif METHOD == 2
      y[i] += -xm*(xm-xl)*(DT/DU); /* left hand derivative */
#elif METHOD == 3
      y[i] += -xm*(xr-xl)*(DT/(2*DU)); /* centred derivative */
#elif METHOD == 4
      y[i] += -xm*(xr-xm)*(DT/DU); /* right derivative ... */
      y[i] -= PARAM_sigma*PARAM_sigma/(4*PARAM_nu) * DT; /* ... + compensation */
#elif METHOD == 5
      y[i] += -(xr*xr-xm*xm)*(.5*DT/DU); /* right hand derivative of u^2/2 */
#elif METHOD == 6
      double x2 = (xr+xm > 0) ? xm*xm : xr*xr;
      double x1 = (xm+xl > 0) ? xl*xl : xm*xm;
      y[i] += -(x2-x1)*(.5*DT/DU); /* upwind scheme */
#else
#error undefined method
#endif

      y[i] += spde_sigma * jvrand_standard_normal();
    }

    /* apply (I - nu theta DT L)^{-1} */
    solve(NU, -gamma, 1+2*gamma, -gamma, y, x);
  }

  for (i=0; i<NU; ++i) {
    printf("%f %f\n", i*DU, x[i]);
  }
  printf("%f %f\n", UMAX, x[0]);

  xfree(y);
  xfree(x);
  return 0;
}
