/* burgers.c - simulations for figure 2
 *
 * Copyright (C) 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "burgers.h"
#include "random4.h"


static  double *x_fe, *x_gal;	/* reference solutions */

static void
worker1(void *data)
{
  if (data == x_fe) {
    wrong_fe(NU/COARSEN, x_fe, COARSEN);
  } else {
    wrong_gal(NU/COARSEN, x_gal, COARSEN);
  }
}


struct work {
  /* input */
  double correction;

  /* results */
  double res_fe, res_gal;
};

static void
worker2(void *data)
{
  struct work *info = data;
  double *x;
  int  i;

  x = xnew(double, NU);

  /* compute the solution for the given correction */
  good_fe(NU, x, info->correction);

  /* compute distance to both targets */
  double res_fe = 0;
  double res_gal = 0;
  for (i=0; i<NU/COARSEN; i+=1) {
    double d;

    d = x_fe[i] - x[i*COARSEN];
    res_fe += d*d;

    d = x_gal[i] - x[i*COARSEN];
    res_gal += d*d;
  }
  info->res_fe = sqrt(res_fe*UMAX/(NU/COARSEN));
  info->res_gal = sqrt(res_gal*UMAX/(NU/COARSEN));

  xfree(x);
}

int
main()
{
  struct factory *f;
  struct work *job;
  const double c_min = .0;
  const double c_max = .5;
  const int steps = 100;
  int  i;

  setpriority(PRIO_PROCESS, 0, PRIO_MAX);

  f = new_factory(0);

  assert(NU%COARSEN == 0);
  x_fe = xnew(double, NU/COARSEN);
  factory_job_start(f, worker1, x_fe);
  x_gal = xnew(double, NU/COARSEN);
  factory_job_start(f, worker1, x_gal);
  factory_job_finish(f, x_fe);
  factory_job_finish(f, x_gal);

  job = xnew(struct work, steps+1);
  for (i=0; i<=steps; ++i) {
    job[i].correction = c_min + (c_max-c_min)/steps*i;
    factory_job_start(f, worker2, job+i);
  }

  for (i=0; i<=steps; ++i) {
    factory_job_finish(f, job+i);
    printf("%f %f %f\n", job[i].correction, job[i].res_fe, job[i].res_gal);
    fflush(NULL);
  }

  xfree(job);
  xfree(x_fe);
  xfree(x_gal);
  delete_factory(f);
  return 0;
}
