/* pool.c - delayed/parallel computation via a thread pool
 *
 * Copyright (C) 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <unistd.h>
#if linux
#include <sched.h>
#endif
#include <pthread.h>
#include <assert.h>

#include "burgers.h"


/**********************************************************************
 * job management
 */

struct job {
  struct job *next;

  int ready_flag;
  pthread_cond_t ready;
  pthread_mutex_t mutex;

  void (*fn)(void *);
  void *data;
};

static struct job *
new_job(void (*fn)(void *), void *data)
{
  struct job *res;

  res = xnew(struct job, 1);
  res->next = NULL;
  res->ready_flag = 0;
  pthread_cond_init(&res->ready, NULL);
  pthread_mutex_init(&res->mutex, NULL);

  res->fn = fn;
  res->data = data;

  return res;
}

static void
delete_job(struct job *job)
{
  xfree(job);
}

/**********************************************************************
 * factories
 */

struct factory {
  int n_workers, n_idle;

  struct job *todo_head, *todo_tail; /* not yet started jobs */
  struct job *done_head, *done_tail;   /* started/completed jobs */
  pthread_cond_t start;		/* work became available */
  pthread_mutex_t mutex;
};

static void *
worker_fn(void *data)
/* process jobs from the todo list */
{
  struct factory *f = data;
  int done = 0;

  pthread_mutex_lock(&f->mutex);
  f->n_workers += 1;
  pthread_mutex_unlock(&f->mutex);

  while (! done) {
    struct job *job;

    pthread_mutex_lock(&f->mutex);

    /* wait for a new job */
    f->n_idle += 1;
    while (! f->todo_head)
      pthread_cond_wait(&f->start, &f->mutex);
    assert(f->n_idle > 0);
    f->n_idle -= 1;

    /* remove job from the todo list */
    job = f->todo_head;
    f->todo_head = job->next;
    if (f->todo_tail == job)  f->todo_tail = NULL;

    /* append job to the done list */
    job->next = NULL;
    if (f->done_head) {
      f->done_tail->next = job;
      f->done_tail = job;
    } else {
      f->done_head = f->done_tail = job;
    }

    pthread_mutex_unlock(&f->mutex);

    /* do the computation */
    if (job->fn) {
      job->fn(job->data);
    } else {
      done = 1;
    }

    /* tell the caller that the result is ready */
    pthread_mutex_lock(&job->mutex);
    assert(job->ready_flag == 0);
    job->ready_flag = 1;
    pthread_cond_signal(&job->ready);
    pthread_mutex_unlock(&job->mutex);
  }

  /* shut down the worker */
  pthread_mutex_lock(&f->mutex);
  f->n_workers -= 1;
  pthread_mutex_unlock(&f->mutex);

  return NULL;
}

static void
factory_job_start_nocheck(struct factory *f, void (*fn)(void *), void *data)
{
  struct job *res;

  res = new_job(fn, data);

  pthread_mutex_lock(&f->mutex);
  if (f->todo_head) {
    f->todo_tail->next = res;
    f->todo_tail = res;
  } else {
    f->todo_head = f->todo_tail = res;
  }
  if (f->n_idle > 0)
    pthread_cond_signal(&f->start);
  pthread_mutex_unlock(&f->mutex);
}

struct factory *
new_factory(int n_workers)
{
  struct factory *res;
  int sys_cpus = sysconf(_SC_NPROCESSORS_CONF);
  int i;

  res = xnew(struct factory, 1);

  res->n_workers = 0;
  res->n_idle = 0;

  res->todo_head = NULL;
  res->todo_tail = NULL;
  res->done_head = NULL;
  res->done_tail = NULL;
  pthread_cond_init(&res->start, NULL);
  pthread_mutex_init(&res->mutex, NULL);

  if (n_workers <= 0)
    n_workers = sys_cpus;
  for (i=0; i<n_workers; ++i) {
    pthread_t thread;
    pthread_create(&thread, NULL, worker_fn, res);
    pthread_detach(thread);
#if linux
    /* Bind each worker to its own CPU.  This is to work around the
     * fact that the automatic scheduling seems to not cope very well
     * with our load.
     */
    if (n_workers <= sys_cpus) {
      cpu_set_t cpu;
      CPU_ZERO(&cpu);
      CPU_SET(i, &cpu);
      pthread_setaffinity_np(thread, sizeof(cpu), &cpu);
    }
#endif
  }

  return res;
}

void
delete_factory(struct factory *f)
{
  for (;;) {
    int done;
    pthread_mutex_lock(&f->mutex);
    done = (f->n_workers == 0);
    pthread_mutex_unlock(&f->mutex);

    if (done)  break;

    factory_job_start_nocheck(f, NULL, NULL);
    factory_job_finish(f, NULL);
  }

  /* free jobs which were not collected */
  while (f->todo_head) {
    struct job *next = f->todo_head->next;
    xfree(f->todo_head);
    f->todo_head = next;
  }
  while (f->done_head) {
    struct job *next = f->done_head->next;
    xfree(f->done_head);
    f->done_head = next;
  }

  xfree(f);
}

void
factory_job_start(struct factory *f, void (*fn)(void *), void *data)
{
  assert(fn != NULL);
  factory_job_start_nocheck(f, fn, data);
}

int
factory_job_finish(struct factory *f, void *data)
{
  struct job *job, *prev;

  /* find the job */
  pthread_mutex_lock(&f->mutex);
  job = f->done_head;
  while (job && job->data != data)
    job = job->next;
  if (! job) {
    prev = NULL;
    job = f->todo_head;
    while (job && job->data != data) {
      prev = job;
      job = job->next;
    }
    if (! job) {
      pthread_mutex_unlock(&f->mutex);
      return -1;
    }
    if (prev) {			/* move job to the head of the queue */
      prev->next = job->next;
      job->next = f->todo_head;
      f->todo_head = job;
      if (f->todo_tail == job)  f->todo_tail = prev;
    }
  }
  pthread_mutex_unlock(&f->mutex);

  /* wait for completion */
  pthread_mutex_lock(&job->mutex);
  if (! job->ready_flag)
    pthread_cond_wait(&job->ready, &job->mutex);
  pthread_mutex_unlock(&job->mutex);

  /* remove the job from the done queue */
  pthread_mutex_lock(&f->mutex);
  prev = NULL;
  job = f->done_head;
  while (job->data != data) {
    prev = job;
    job = job->next;
  }
  if (prev) {
    prev->next = job->next;
  } else {
    f->done_head = job->next;
  }
  if (f->done_tail == job)  f->done_tail = prev;
  pthread_mutex_unlock(&f->mutex);

  delete_job(job);
  return 0;
}

#ifdef TESTING
#include <stdio.h>

static void
job_fn(void *data)
{
  int *ptr = data;
  int  i = *ptr;
  usleep((i%101 == 7)*100000 + 10000);
  *ptr += 1;
}

int
main()
{
  struct factory *f = new_factory(0);
  int  data[503];
  int  i;

  for (i=0; i<503; ++i) {
    data[i] = i;
    factory_job_start(f, job_fn, data+i);
  }
  for (i=0; i<503; ++i) {
    int  j = ((i+1)*37) % 503;	/* all the jobs in "random" order */
    factory_job_finish(f, data+j);
    if (data[j] != j+1) {
      fprintf(stderr, "error: got %d instead of %d\n", data[j], j+1);
    }
  }
  fputs("tests done\n", stderr);

  delete_factory(f);
  return 0;
}
#endif
