/* wrong_gal.c - right-handed Galerkin solution
 *
 * Copyright (C) 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#include <assert.h>

#include <fftw3.h>

#include "burgers.h"
#include "random4.h"

void
wrong_gal(int n, double *xx, int l)
{
  struct jvrand *rng;
  double *x, *y;
  double du = UMAX / n;
  const double  spde_sigma = PARAM_sigma * sqrt(DT/(l*du));
  fftw_complex *xhat;
  fftw_plan to, fro;
  unsigned long int  k, kk;
  int  i, j;

  rng = new_jvrand(SEED);

  /* use fftw_malloc to make sure memory alignment is ok */
  x = fftw_malloc(n*sizeof(double));
  for (i=0; i<n; ++i)  x[i] = 0;
  y = fftw_malloc(n*sizeof(double));
  xhat = fftw_malloc((n/2+1)*sizeof(fftw_complex));

  to = fftw_plan_dft_r2c_1d(n, x, xhat, FFTW_DESTROY_INPUT|FFTW_PATIENT);
  fro = fftw_plan_dft_c2r_1d(n, xhat, x, FFTW_DESTROY_INPUT|FFTW_PATIENT);
  /* Applying first 'to' and then 'fro' multiplies 'x' by the factor 'n'.
   * To compensate, we divide by n every time after we applied 'fro'.  */

  /* main loop */
  kk = TMAX/DT+0.5;
  for (k=0; k<kk; ++k) {
    double  xl, xm, xr;

    /* apply (I + nu (1-theta) DT L) */
    fftw_execute_dft_r2c(to, x, xhat);
    for (i=0; i<n/2+1; ++i) {
      double q = 1 - (PARAM_nu*(1-THETA)*DT)*(i*i);
      xhat[i][0] *= q;
      xhat[i][1] *= q;
    }
    fftw_execute_dft_c2r(fro, xhat, y);
    for (i=0; i<n; ++i) y[i] *= (1.0/n);

    /* add drift and noise */
    xm = x[n-1], xr = x[0];
    for (i=0; i<n; ++i) {
      xl = xm, xm = xr, xr = (i+1<n) ? x[i+1] : x[0];
      y[i] += -xm*(xr-xm)/du * DT; /* right hand derivative */

      double noise = 0;
      for (j=0; j<l; ++j) {
	/* We need all l calls to jvrand_standard_normal(), in order
	 * to get the same noise as for the fine grid. */
	noise += jvrand_standard_normal(rng);
      }
      y[i] += spde_sigma * noise;
    }

    /* apply (I - nu theta DT L)^{-1} */
    fftw_execute_dft_r2c(to, y, xhat);
    for (i=0; i<n/2+1; ++i) {
      double q = 1.0 / (1 + (PARAM_nu*THETA*DT)*(i*i));
      xhat[i][0] *= q;
      xhat[i][1] *= q;
    }
    fftw_execute_dft_c2r(fro, xhat, x);
    for (i=0; i<n; ++i) x[i] *= (1.0/n);
  }

  /* copy back the data */
  for (i=0; i<n; ++i)  xx[i] = x[i];

  fftw_destroy_plan(to);
  fftw_destroy_plan(fro);
  fftw_free(xhat);
  fftw_free(y);
  fftw_free(x);

  fftw_cleanup();

  delete_jvrand(rng);
}
