/* good_fe.c - known-good finite elements discretisation (with additional drift)
 *
 * Copyright (C) 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#include <math.h>
#include <assert.h>

#include "burgers.h"
#include "random4.h"


void
good_fe(int n, double *x, double correction)
{
  struct jvrand *rng;
  double *y;
  double du = UMAX / n;
  const double  gamma = THETA*DT*PARAM_nu/(du*du);
  const double  delta = (1-THETA)*DT*PARAM_nu/(du*du);
  const double  spde_sigma = PARAM_sigma * sqrt(DT/du);
  unsigned long int  k, kk;
  int  i;

  assert(5*DT < du);	    /* plausibility check for CFL condition */

  rng = new_jvrand(SEED);
  y = xnew(double, n);
  for (i=0; i<n; ++i)  x[i] = 0;

  /* main loop */
  kk = TMAX/DT+0.5;
  for (k=0; k<kk; ++k) {
    double  xl, xm, xr;

    /* apply (I + nu (1-theta) DT L) */
    xmul(n, delta, 1-2*delta, delta, x, y);

    /* add drift and noise */
    xm = x[n-1], xr = x[0];
    for (i=0; i<n; ++i) {
      xl = xm, xm = xr, xr = (i+1<n) ? x[i+1] : x[0];
      y[i] += -.5*(xm*xm-xl*xl)/du * DT; /* left hand derivative of u^2/2 */
      y[i] += correction * (PARAM_sigma*PARAM_sigma/PARAM_nu*DT); /* ... + compensation */
      y[i] += spde_sigma * jvrand_standard_normal(rng);
    }

    /* apply (I - nu theta DT L)^{-1} */
    solve(n, -gamma, 1+2*gamma, -gamma, y, x);
  }

  xfree(y);
  delete_jvrand(rng);
}
