/* random4.c - random number generation
 *
 * This file combines the SFMT-19937 random number generator by
 * Mutsuo Saito and Makoto Matsumoto with the Ziggurat method as
 * described by George Marsaglia and Wai Wan Tsang.
 *
 * The Mersenne Twister code "SFMT-src-1.3.3.tar.gz" was downloaded
 * from the SFMT home page at
 * http://www.math.sci.hiroshima-u.ac.jp/~m-mat/MT/SFMT/index.html
 * and heavily edited.  The code from SFMT is Copyright (C) 2006, 2007
 * Mutsuo Saito and Makoto Matsumoto and can be distributed under the
 * following BSD-style license:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *         * Redistributions of source code must retain the above
 *           copyright notice, this list of conditions and the
 *           following disclaimer.
 *         * Redistributions in binary form must reproduce the above
 *           copyright notice, this list of conditions and the
 *           following disclaimer in the documentation and/or other
 *           materials provided with the distribution.
 *         * Neither the name of the Hiroshima University nor the
 *           names of its contributors may be used to endorse or
 *           promote products derived from this software without
 *           specific prior written permission.
 *
 *     THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 *     CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 *     INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *     DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *     CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *     NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *     LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *     HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *     OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *     EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * All other code in this file is
 * Copyright (C) 2010 Jochen Voss
 * and can be distributed under the GPL:
 *
 *     This program is free software; you can redistribute it and/or
 *     modify it under the terms of the GNU General Public License as
 *     published by the Free Software Foundation; either version 2 of
 *     the License, or (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public
 *     License along with this program; if not, write to the Free
 *     Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 *     MA 02110-1301 USA
 *
 * Please send all comments, questions, bug reports, etc. about this
 * code to Jochen Voss <voss@seehuhn.de>.
 */

#include <stdlib.h>
#include <math.h>
#include <assert.h>

#include "random4.h"

/* parameters */
#define MEXP    19937
#define N       (MEXP / 128 + 1)
#define POS1	122
#define SL1	18
#define SL2	1
#define SR1	11
#define SR2	1
#define MSK1	0xdfffffefU
#define MSK2	0xddfecb7fU
#define MSK3	0xbffaffffU
#define MSK4	0xbffffff6U
#define PARITY1	0x00000001U
#define PARITY2	0x00000000U
#define PARITY3	0x00000000U
#define PARITY4	0x13c9e684U

#if __SSE2__
#include <emmintrin.h>
union W128_T {
  __m128i si;
  uint32_t u[4];
};
#else
union W128_T {
  uint32_t u[4];
};
#endif


struct jvrand {			/* RNG state */
  union W128_T  sfmt[N];
  uint32_t *psfmt32;
  int  idx;
  uint32_t parity[4];
};


#if __SSE2__

static inline __m128i
mm_recursion(__m128i *a, __m128i *b, __m128i c, __m128i d, __m128i mask)
{
  __m128i v, x, y, z;

  x = _mm_load_si128(a);
  y = _mm_srli_epi32(*b, SR1);
  z = _mm_srli_si128(c, SR2);
  v = _mm_slli_epi32(d, SL1);
  z = _mm_xor_si128(z, x);
  z = _mm_xor_si128(z, v);
  x = _mm_slli_si128(x, SL2);
  y = _mm_and_si128(y, mask);
  z = _mm_xor_si128(z, x);
  z = _mm_xor_si128(z, y);
  return z;
}

static inline void
gen_rand_all(struct jvrand *rng)
{
  union W128_T *sfmt = rng->sfmt;
  __m128i r, r1, r2, mask;
  int i;

  mask = _mm_set_epi32(MSK4, MSK3, MSK2, MSK1);

  r1 = _mm_load_si128(&sfmt[N - 2].si);
  r2 = _mm_load_si128(&sfmt[N - 1].si);
  for (i = 0; i < N - POS1; i++) {
    r = mm_recursion(&sfmt[i].si, &sfmt[i + POS1].si, r1, r2, mask);
    _mm_store_si128(&sfmt[i].si, r);
    r1 = r2;
    r2 = r;
  }
  for (; i < N; i++) {
    r = mm_recursion(&sfmt[i].si, &sfmt[i + POS1 - N].si, r1, r2, mask);
    _mm_store_si128(&sfmt[i].si, r);
    r1 = r2;
    r2 = r;
  }
}

#else

static inline void
rshift128(union W128_T *out, union W128_T const *in, int shift)
{
  uint64_t th, tl, oh, ol;

  th = ((uint64_t) in->u[3] << 32) | ((uint64_t) in->u[2]);
  tl = ((uint64_t) in->u[1] << 32) | ((uint64_t) in->u[0]);

  oh = th >> (shift * 8);
  ol = tl >> (shift * 8);
  ol |= th << (64 - shift * 8);
  out->u[1] = (uint32_t) (ol >> 32);
  out->u[0] = (uint32_t) ol;
  out->u[3] = (uint32_t) (oh >> 32);
  out->u[2] = (uint32_t) oh;
}

static inline void
lshift128(union W128_T *out, union W128_T const *in, int shift)
{
  uint64_t th, tl, oh, ol;

  th = ((uint64_t) in->u[3] << 32) | ((uint64_t) in->u[2]);
  tl = ((uint64_t) in->u[1] << 32) | ((uint64_t) in->u[0]);

  oh = th << (shift * 8);
  ol = tl << (shift * 8);
  oh |= tl >> (64 - shift * 8);
  out->u[1] = (uint32_t) (ol >> 32);
  out->u[0] = (uint32_t) ol;
  out->u[3] = (uint32_t) (oh >> 32);
  out->u[2] = (uint32_t) oh;
}

static inline void
do_recursion(union W128_T *r, union W128_T *a, union W128_T *b, union W128_T *c, union W128_T *d)
{
  union W128_T x;
  union W128_T y;

  lshift128(&x, a, SL2);
  rshift128(&y, c, SR2);
  r->u[0] = a->u[0] ^ x.u[0] ^ ((b->u[0] >> SR1) & MSK1) ^ y.u[0] ^ (d->u[0] << SL1);
  r->u[1] = a->u[1] ^ x.u[1] ^ ((b->u[1] >> SR1) & MSK2) ^ y.u[1] ^ (d->u[1] << SL1);
  r->u[2] = a->u[2] ^ x.u[2] ^ ((b->u[2] >> SR1) & MSK3) ^ y.u[2] ^ (d->u[2] << SL1);
  r->u[3] = a->u[3] ^ x.u[3] ^ ((b->u[3] >> SR1) & MSK4) ^ y.u[3] ^ (d->u[3] << SL1);
}

static inline void
gen_rand_all(struct jvrand *rng)
{
  union W128_T *sfmt = rng->sfmt;
  union W128_T *r1, *r2;
  int i;

  r1 = &sfmt[N - 2];
  r2 = &sfmt[N - 1];
  for (i = 0; i < N - POS1; i++) {
    do_recursion(&sfmt[i], &sfmt[i], &sfmt[i + POS1], r1, r2);
    r1 = r2;
    r2 = &sfmt[i];
  }
  for (; i < N; i++) {
    do_recursion(&sfmt[i], &sfmt[i], &sfmt[i + POS1 - N], r1, r2);
    r1 = r2;
    r2 = &sfmt[i];
  }
}
#endif

static void
period_certification(struct jvrand *rng)
{
  uint32_t *psfmt32 = rng->psfmt32;
  uint32_t *parity = rng->parity;
  int inner = 0;
  int i, j;
  uint32_t work;

  for (i = 0; i < 4; i++)
    inner ^= psfmt32[i] & parity[i];
  for (i = 16; i > 0; i >>= 1)
    inner ^= inner >> i;
  inner &= 1;
  if (inner == 1) return;
  for (i = 0; i < 4; i++) {
    work = 1;
    for (j = 0; j < 32; j++) {
      if ((work & parity[i]) != 0) {
	psfmt32[i] ^= work;
	return;
      }
      work = work << 1;
    }
  }
}

struct jvrand *
new_jvrand(uint32_t seed)
{
  struct jvrand *res;
  uint32_t *psfmt32;
  int i;

  res = malloc(sizeof(struct jvrand));
  if (! res)  return NULL;

  res->psfmt32 = psfmt32 = &res->sfmt[0].u[0];
  res->parity[0] = PARITY1;
  res->parity[1] = PARITY2;
  res->parity[2] = PARITY3;
  res->parity[3] = PARITY4;

  psfmt32[0] = seed;
  for (i = 1; i < 4*N; i++) {
    psfmt32[i] = 1812433253UL * (psfmt32[i - 1] ^ (psfmt32[i - 1] >> 30)) + i;
  }
  res->idx = 4*N;
  period_certification(res);

  return res;
}

void
delete_jvrand(struct jvrand *rng)
{
  free(rng);
}

static inline uint32_t
jvrand_int32(struct jvrand *rng)
{
  uint32_t r;

  if (rng->idx >= 4*N) {
    gen_rand_all(rng);
    rng->idx = 0;
  }
  r = rng->psfmt32[rng->idx++];
  return r;
}

double
jvrand_uniform(struct jvrand *rng)
/* generates a random number from the interval [0,1).  */
{
  return  jvrand_int32(rng) * (1.0/4294967296.0); /* divided by 2^32 */
}

/* number of levels in the ziggurat */
#define LEVELS 128

/* position of right-most step */
#define PARAM_R 3.44428647676

/* level values */
static const double ytab[128] = {
  1, 0.963598623011, 0.936280813353, 0.913041104253,
  0.892278506696, 0.873239356919, 0.855496407634, 0.838778928349,
  0.822902083699, 0.807732738234, 0.793171045519, 0.779139726505,
  0.765577436082, 0.752434456248, 0.739669787677, 0.727249120285,
  0.715143377413, 0.703327646455, 0.691780377035, 0.68048276891,
  0.669418297233, 0.65857233912, 0.647931876189, 0.637485254896,
  0.62722199145, 0.617132611532, 0.607208517467, 0.597441877296,
  0.587825531465, 0.578352913803, 0.569017984198, 0.559815170911,
  0.550739320877, 0.541785656682, 0.532949739145, 0.524227434628,
  0.515614886373, 0.507108489253, 0.498704867478, 0.490400854812,
  0.482193476986, 0.47407993601, 0.466057596125, 0.458123971214,
  0.450276713467, 0.442513603171, 0.434832539473, 0.427231532022,
  0.419708693379, 0.41226223212, 0.404890446548, 0.397591718955,
  0.390364510382, 0.383207355816, 0.376118859788, 0.369097692334,
  0.362142585282, 0.355252328834, 0.348425768415, 0.341661801776,
  0.334959376311, 0.328317486588, 0.321735172063, 0.31521151497,
  0.308745638367, 0.302336704338, 0.29598391232, 0.289686497571,
  0.283443729739, 0.27725491156, 0.271119377649, 0.265036493387,
  0.259005653912, 0.253026283183, 0.247097833139, 0.241219782932,
  0.235391638239, 0.229612930649, 0.223883217122, 0.218202079518,
  0.212569124201, 0.206983981709, 0.201446306496, 0.195955776745,
  0.190512094256, 0.185114984406, 0.179764196185, 0.174459502324,
  0.169200699492, 0.1639876086, 0.158820075195, 0.153697969964,
  0.148621189348, 0.143589656295, 0.138603321143, 0.133662162669,
  0.128766189309, 0.123915440582, 0.119109988745, 0.114349940703,
  0.10963544023, 0.104966670533, 0.100343857232, 0.0957672718266,
  0.0912372357329, 0.0867541250127, 0.082318375932, 0.0779304915295,
  0.0735910494266, 0.0693007111742, 0.065060233529, 0.0608704821745,
  0.056732448584, 0.05264727098, 0.0486162607163, 0.0446409359769,
  0.0407230655415, 0.0368647267386, 0.0330683839378, 0.0293369977411,
  0.0256741818288, 0.0220844372634, 0.0185735200577, 0.0151490552854,
  0.0118216532614, 0.00860719483079, 0.00553245272614, 0.00265435214565
};

/* quick acceptance check */
static const uint32_t ktab[128] = {
  0, 12590644, 14272653, 14988939,
  15384584, 15635009, 15807561, 15933577,
  16029594, 16105155, 16166147, 16216399,
  16258508, 16294295, 16325078, 16351831,
  16375291, 16396026, 16414479, 16431002,
  16445880, 16459343, 16471578, 16482744,
  16492970, 16502368, 16511031, 16519039,
  16526459, 16533352, 16539769, 16545755,
  16551348, 16556584, 16561493, 16566101,
  16570433, 16574511, 16578353, 16581977,
  16585398, 16588629, 16591685, 16594575,
  16597311, 16599901, 16602354, 16604679,
  16606881, 16608968, 16610945, 16612818,
  16614592, 16616272, 16617861, 16619363,
  16620782, 16622121, 16623383, 16624570,
  16625685, 16626730, 16627708, 16628619,
  16629465, 16630248, 16630969, 16631628,
  16632228, 16632768, 16633248, 16633671,
  16634034, 16634340, 16634586, 16634774,
  16634903, 16634972, 16634980, 16634926,
  16634810, 16634628, 16634381, 16634066,
  16633680, 16633222, 16632688, 16632075,
  16631380, 16630598, 16629726, 16628757,
  16627686, 16626507, 16625212, 16623794,
  16622243, 16620548, 16618698, 16616679,
  16614476, 16612071, 16609444, 16606571,
  16603425, 16599973, 16596178, 16591995,
  16587369, 16582237, 16576520, 16570120,
  16562917, 16554758, 16545450, 16534739,
  16522287, 16507638, 16490152, 16468907,
  16442518, 16408804, 16364095, 16301683,
  16207738, 16047994, 15704248, 15472926
};

/* quick value conversion */
static const double wtab[128] = {
  1.62318314817e-08, 2.16291505214e-08, 2.54246305087e-08, 2.84579525938e-08,
  3.10340022482e-08, 3.33011726243e-08, 3.53439060345e-08, 3.72152672658e-08,
  3.8950989572e-08, 4.05763964764e-08, 4.21101548915e-08, 4.35664624904e-08,
  4.49563968336e-08, 4.62887864029e-08, 4.75707945735e-08, 4.88083237257e-08,
  5.00063025384e-08, 5.11688950428e-08, 5.22996558616e-08, 5.34016475624e-08,
  5.44775307871e-08, 5.55296344581e-08, 5.65600111659e-08, 5.75704813695e-08,
  5.85626690412e-08, 5.95380306862e-08, 6.04978791776e-08, 6.14434034901e-08,
  6.23756851626e-08, 6.32957121259e-08, 6.42043903937e-08, 6.51025540077e-08,
  6.59909735447e-08, 6.68703634341e-08, 6.77413882848e-08, 6.8604668381e-08,
  6.94607844804e-08, 7.03102820203e-08, 7.11536748229e-08, 7.1991448372e-08,
  7.2824062723e-08, 7.36519550992e-08, 7.44755422158e-08, 7.52952223703e-08,
  7.61113773308e-08, 7.69243740467e-08, 7.77345662086e-08, 7.85422956743e-08,
  7.93478937793e-08, 8.01516825471e-08, 8.09539758128e-08, 8.17550802699e-08,
  8.25552964535e-08, 8.33549196661e-08, 8.41542408569e-08, 8.49535474601e-08,
  8.57531242006e-08, 8.65532538723e-08, 8.73542180955e-08, 8.8156298059e-08,
  8.89597752521e-08, 8.97649321908e-08, 9.05720531451e-08, 9.138142487e-08,
  9.21933373471e-08, 9.30080845407e-08, 9.38259651738e-08, 9.46472835298e-08,
  9.54723502847e-08, 9.63014833769e-08, 9.71350089201e-08, 9.79732621669e-08,
  9.88165885297e-08, 9.96653446693e-08, 1.00519899658e-07, 1.0138063623e-07,
  1.02247952126e-07, 1.03122261554e-07, 1.04003996769e-07, 1.04893609795e-07,
  1.05791574313e-07, 1.06698387725e-07, 1.07614573423e-07, 1.08540683296e-07,
  1.09477300508e-07, 1.1042504257e-07, 1.11384564771e-07, 1.12356564007e-07,
  1.13341783071e-07, 1.14341015475e-07, 1.15355110887e-07, 1.16384981291e-07,
  1.17431607977e-07, 1.18496049514e-07, 1.19579450872e-07, 1.20683053909e-07,
  1.21808209468e-07, 1.2295639141e-07, 1.24129212952e-07, 1.25328445797e-07,
  1.26556042658e-07, 1.27814163916e-07, 1.29105209375e-07, 1.30431856341e-07,
  1.31797105598e-07, 1.3320433736e-07, 1.34657379914e-07, 1.36160594606e-07,
  1.37718982103e-07, 1.39338316679e-07, 1.41025317971e-07, 1.42787873535e-07,
  1.44635331499e-07, 1.4657889173e-07, 1.48632138436e-07, 1.50811780719e-07,
  1.53138707402e-07, 1.55639532047e-07, 1.58348931426e-07, 1.61313325908e-07,
  1.64596952856e-07, 1.68292495203e-07, 1.72541128694e-07, 1.77574279496e-07,
  1.83813550477e-07, 1.92166040885e-07, 2.05295471952e-07, 2.22600839893e-07
};

double
jvrand_standard_normal(struct jvrand *rng)
/* generate a standard Gaussian random number */
{
  uint32_t  U, sigma, i, j;
  double  x, y;

  while (1) {
    U = jvrand_int32(rng);
    i = U & 0x0000007F;
    sigma = U & 0x00000080;
    j = U>>8;

    x = j*wtab[i];
    if (j < ktab[i])  break;

    if (i<LEVELS-1) {
      double  y0, y1;
      y0 = ytab[i];
      y1 = ytab[i+1];
      y = y1+(y0-y1)*jvrand_uniform(rng);
    } else {
      x = PARAM_R - log(1.0-jvrand_uniform(rng))/PARAM_R;
      y = exp(-PARAM_R*(x-0.5*PARAM_R))*jvrand_uniform(rng);
    }
    if (y < exp(-0.5*x*x))  break;
  }
  return  sigma ? x : -x;
}

#ifdef TESTING
int
main()
{
  struct jvrand *rng;

  rng = new_jvrand(12345);

  assert(jvrand_int32(rng) == 1811630862);
  assert(jvrand_int32(rng) == 4267490791);
  assert(jvrand_int32(rng) == 116883287);
  assert(jvrand_int32(rng) == 2178177053);
  assert(jvrand_int32(rng) == 1355292313);
  assert(jvrand_int32(rng) == 293457464);
  assert(jvrand_int32(rng) == 3566685862);
  assert(jvrand_int32(rng) == 574122430);
  assert(jvrand_int32(rng) == 1545067006);
  assert(jvrand_int32(rng) == 3489303066);

  delete_jvrand(rng);
  return 0;
}
#endif
