/* direct.c - solve cyclic tridiagonal systems using Gauss algorithm
 *
 * Copyright (C) 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#if __SSE2__
#include <emmintrin.h>
#endif

#include "strange.h"

void
xmul(int n, double Al, double Am, double Ar,
     const double *restrict in, double *restrict out)
/* Left-multiply 'in' with the cyclic tri-band matrix which has 'Al'
 * on the sub-diagonal, 'Am' on the diagonal and 'Ar' on the
 * super-diagonal.  Store the result in 'out'.  */
{
#if __SSE2__
  __m128d aa, bb, tmp, yy, ll, mm, rr;
  double xl, xm, xr;
  int i;

  xl = in[n-1], xm = in[0], xr = in[1];
  out[0] = Al*xl + Am*xm + Ar*xr;
  i = 1;

  xl = xm, xm = xr, xr=in[i+1];
  out[i] = Al*xl + Am*xm + Ar*xr;
  ++i;

  ll = _mm_set1_pd(Al);
  mm = _mm_set1_pd(Am);
  rr = _mm_set1_pd(Ar);

  aa = _mm_setr_pd(xm, xr);
  bb = _mm_setr_pd(xr, in[3]);
  for (; i+3<n; i += 2) {
    /* precond: aa = (in[i-1], in[i]) and bb = (in[i], in[i+1]) */
    yy = _mm_mul_pd(ll, aa);	 /* yy = ll*aa */
    tmp = _mm_mul_pd(mm, bb);
    yy = _mm_add_pd(yy, tmp);	 /* yy += mm*bb */
    tmp = bb;
    bb = _mm_load_pd(in+i+2);
    /* ... bb = (in[i+2],in[i+3]) */
    aa = _mm_shuffle_pd(tmp, bb, 1);
    /* ... aa = (in[i+1], in[i+2]) */
    tmp = _mm_mul_pd(rr, aa);
    yy = _mm_add_pd(yy, tmp);	 /* yy += rr*aa */
    _mm_store_pd(out+i, yy);
    /* postcond: aa = (in[i+1], in[i+2]) and bb = (in[i+2], in[i+3]) */
  }
  /* now i is either n-3 or n-2 */

  _mm_store_sd(&xl, aa);
  _mm_store_sd(&xm, bb);
  _mm_storeh_pd(&xr, bb);
  out[i] = Al*xl + Am*xm + Ar*xr;
  ++i;
  /* now i is either n-2 or n-1 */

  if (i < n-1) {
    xl = xm, xm = xr, xr=in[i+1];
    out[i] = Al*xl + Am*xm + Ar*xr;
    ++i;
  }

  xl = xm, xm = xr, xr = in[0];
  out[n-1] = Al*xl + Am*xm + Ar*xr;
#else
  double xl, xm, xr;
  int i;

  xl = in[n-1], xm = in[0], xr = in[1];
  out[0] = Al*xl + Am*xm + Ar*xr;
  for (i=1; i<n-1; ++i) {
    xl = xm, xm = xr, xr=in[i+1];
    out[i] = Al*xl + Am*xm + Ar*xr;
  }
  xl = xm, xm = xr, xr = in[0];
  out[n-1] = Al*xl + Am*xm + Ar*xr;
#endif
}

void
solve(int n, double Al, double Am, double Ar,
      const double *restrict rhs, double *restrict x)
/* Solve a system of linear equations using forward/back-substitution.  */
{
  double *y, *z;
  double q, tmp;
  int  i;

  z = xnew(double, n);
  y = xnew(double, n-1);

  /* step 1: eliminate the subdiagonal */
  z[0] = Am;
  y[0] = Al;
  x[0] = rhs[0];
  for (i=1; i<n-2; ++i) {
    q = Al / z[i-1];		/* subtract q times row[i-1] from row[i] */
    z[i] = Am - q*Ar;
    y[i] = - q*y[i-1];
    x[i] = rhs[i] - q*x[i-1];
  }
  q = Al / z[n-3];
  z[n-2] = Am - q*Ar;
  y[n-2] = Ar - q*y[n-3];
  x[n-2] = rhs[n-2] - q*x[n-3];
  q = Al / z[n-2];
  z[n-1] = Am - q*y[n-2];
  x[n-1] = rhs[n-1] - q*x[n-2];

  /* step 2: clear the bottom line */
  tmp = Ar;
  for (i=0; i<n-2; ++i) {
    q = tmp/z[i];		/* subtract q times row[i] from row[n-1] */
    tmp = -q*Ar;		/* the new value in row n-1, col i+1 */
    z[n-1] -= q*y[i];
    x[n-1] -= q*x[i];
  }
  q = tmp/z[n-2];
  z[n-1] -= q*y[n-2];
  x[n-1] -= q*x[n-2];

  /* step 3: solve the system backwards */
  x[n-1] /= z[n-1];
  x[n-2] = (x[n-2] - y[n-2]*x[n-1]) / z[n-2];
  for (i=n-3; i>=0; --i) {
    x[i] = (x[i] - Ar*x[i+1] - y[i]*x[n-1]) / z[i];
  }

  xfree(y);
  xfree(z);
}

#if TESTING
#include <stdio.h>

#include "random4.h"

static double
ddot(int n, const double *x, const double *y)
/* Compute the inner product between 'x' and 'y'.  */
{
#if __SSE2__
  __m128d rr, xx, yy;
  double res[2];
  int i;

  rr = _mm_setzero_pd();
  for (i=0; i+1<n; i+=2) {
    xx = _mm_load_pd(x+i);	/* load two elements from x, ... */
    yy = _mm_load_pd(y+i);	/* load two elements from y, ... */
    xx = _mm_mul_pd(xx, yy);	/* multiply both, and ... */
    rr = _mm_add_pd(rr, xx);	/* add them to the result.  */
  }
  _mm_store_pd(res, rr);
  if (i < n) {
    res[0] += x[i]*y[i];
  }
  return res[0] + res[1];
#else
  double res;
  int i;

  res = 0;
  for (i=0; i<n; ++i) {
    res += x[i] * y[i];
  }
  return res;
#endif
}

static double
residual(int n, double Al, double Am, double Ar,
	 const double *rhs, double *x)
{
  double *y;
  double  res;
  int  i;

  y = xnew(double, n);
  xmul(n, Al, Am, Ar, x, y);
#if __SSE2__
  for (i=0; i+1<n; i+=2) {
    __m128d aa, bb;
    aa = _mm_load_pd(y+i);
    bb = _mm_load_pd(rhs+i);
    aa = _mm_sub_pd(aa, bb);
    _mm_store_pd(y+i, aa);
  }
  if (i<n) y[n-1] -= rhs[n-1];
#else
  for (i=0; i<n; ++i)  y[i] -= rhs[i];
#endif
  res = ddot(n, y, y);
  xfree(y);
  return res;
}

int
main()
{
  struct jvrand *rng = new_jvrand(123);
  int n;
  for (n=4; n<10000; n+=71) {
    double *x, *rhs;
    int j;

    x = xnew(double, n);
    rhs = xnew(double, n);

    for (j=0; j<10; ++j) {
      double Al, Am, Ar;
      int  i;

      for (i=0; i<n; ++i) {
	rhs[i] = jvrand_standard_normal(rng);
      }
      Al = 0.1*jvrand_standard_normal(rng);
      Ar = 0.1*jvrand_standard_normal(rng);
      do {
	Am = 1 + 0.1*jvrand_standard_normal(rng);
      } while (Am <= fabs(Al)+fabs(Ar)+1e-6);

      solve(n, Al, Am, Ar, rhs, x);
      if (j == 0)
	printf("%d %g\n", n, residual(n, Al, Am, Ar, rhs, x));
    }

    xfree(rhs);
    xfree(x);
  }
  delete_jvrand(rng);
  return 0;
}
#endif
