/* spde.c - some finite differences discretisations of the strange SPDE
 *
 * Copyright (C) 2009, 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <math.h>
#include <assert.h>

#include "strange.h"
#include "random4.h"


#if defined(__APPLE__)
static inline void
sincos(double x, double *sin_ret, double *cos_ret)
{
  *sin_ret = sin(x);
  *cos_ret = cos(x);
}
#endif


int
main()
{
  struct jvrand *rng;
  double *x1, *x2, *y1, *y2;
  const double  gamma = THETA*DT*PARAM_nu/(DU*DU);
  const double  delta = (1-THETA)*DT*PARAM_nu/(DU*DU);
  const double  spde_sigma = PARAM_sigma * sqrt(DT/DU);
  unsigned long int  k, kk;
  int  i;

  assert(5*DT < DU);	    /* plausibility check for CFL condition */

  x1 = xnew(double, NU);
  x2 = xnew(double, NU);
  y1 = xnew(double, NU);
  y2 = xnew(double, NU);

  /* initialisation */
  rng = new_jvrand(SEED);
  for (i=0; i<NU; ++i)  x1[i] = sin(i*DU);
  for (i=0; i<NU; ++i)  x2[i] = cos(i*DU);

  /* main loop */
  kk = TMAX/DT+0.5;
  for (k=0; k<kk; ++k) {
    double  x1l, x1m, x1r;
    double  x2l, x2m, x2r;

    /* apply (I + nu (1-theta) DT L) */
    xmul(NU, delta, 1-2*delta, delta, x1, y1);
    xmul(NU, delta, 1-2*delta, delta, x2, y2);

    /* add drift and noise */
    x1m = x1[NU-1], x1r = x1[0];
    x2m = x2[NU-1], x2r = x2[0];
    for (i=0; i<NU; ++i) {
      double  dx1, dx2;

      x1l = x1m, x1m = x1r, x1r = (i+1<NU) ? x1[i+1] : x1[0];
      x2l = x2m, x2m = x2r, x2r = (i+1<NU) ? x2[i+1] : x2[0];
#if METHOD == 1
      dx1 = (x1r-x1m)*(DT/DU);	/* right hand derivative */
      dx2 = (x2r-x2m)*(DT/DU);
#elif METHOD == 2
      dx1 = (x1r-x1l)*(DT/(2*DU)); /* centred derivative */
      dx2 = (x2r-x2l)*(DT/(2*DU));
#elif METHOD == 3
      dx1 = (x1m-x1l)*(DT/DU);	/* left hand derivative */
      dx2 = (x2m-x2l)*(DT/DU);
#elif METHOD == 4
      dx1 = (x1r-x1l)*(DT/(2*DU)); /* centred derivative + ... */
      dx2 = (x2r-x2l)*(DT/(2*DU));
      y1[i] += sin(x2m) * DT; /* ... + compensation */
      y2[i] += -cos(x1m) * DT;
#else
#error undefined method
#endif

      double s1, c1, s2, c2;
      sincos(x1m, &s1, &c1);
      sincos(x2m, &s2, &c2);
      double q1 = 2.0/(SDE_SIGMA*SDE_SIGMA);
      double q2 = 4.0/(SDE_SIGMA*SDE_SIGMA);
      y1[i] += q1*(c2-s1)*dx2 + q2*s1*c1*DT;
      y2[i] += q1*(s1-c2)*dx1 - q2*c2*DT;

      y1[i] += spde_sigma*jvrand_standard_normal(rng);
      y2[i] += spde_sigma*jvrand_standard_normal(rng);
    }

    /* apply (I - nu theta DT L)^{-1} */
    solve(NU, -gamma, 1+2*gamma, -gamma, y1, x1);
    solve(NU, -gamma, 1+2*gamma, -gamma, y2, x2);
  }

  for (i=0; i<NU; ++i) {
    printf("%f %f %f\n", i*DU, x1[i], x2[i]);
  }

  xfree(y1);
  xfree(y2);
  xfree(x1);
  xfree(x2);
  delete_jvrand(rng);
  return 0;
}
