/* burgers.c - finite differences discretisation of Burgers' SPDE
 *
 * Copyright (C) 2009, 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>
#include <float.h>

#include "burgers.h"
#include "random4.h"


static  double *x_fe;		/* reference solution */


int
try_read(const char *execname, int n, double *res)
{
  time_t  t1, t2, t3;
  struct stat  sbuf;
  int  rc;

  rc = stat(execname, &sbuf);
  if (rc < 0) {
    fprintf(stderr, "warning: cannot stat \"%s\" (%m)\n", execname);
    return -1;
  }
  t1 = sbuf.st_mtime;

  rc = stat("path.dat", &sbuf);
  if (rc < 0)  return -1;
  t2 = sbuf.st_mtime;
  if (difftime(t2, t1) < 0)  return -1;

  rc = stat("best", &sbuf);
  if (rc < 0)  return -1;
  t3 = sbuf.st_mtime;
  if (difftime(t3, t1) < 0)  return -1;

  fprintf(stderr, "reading previous state from \"best\"\n");

  FILE *fp = fopen("best", "r");
  int  i;

  if (fscanf(fp, "# %*f\n") < 0)  return -1;
  for (i=0; i<=n; ++i) {
    int k, rc;
    rc = fscanf(fp, "%lf*x**%d", res+i, &k);
    if (rc != 2 || k!=i) return -1;
  }
  return 0;
}

static double
corrected(const double *coeff, void *data)
{
  double *x;
  int  i;

  x = xnew(double, NU);

  /* compute the solution for the given correction */
  good_fe(NU, x, DEGREE, coeff);

  double res_fe;
  if (isnan(x[0])) {
    res_fe = DBL_MAX;
  } else {
    res_fe = 0;
    for (i=0; i<NU/COARSEN; i+=1) {
      double d = x_fe[i] - x[i*COARSEN];
      res_fe += d*d;
    }
    res_fe = sqrt(res_fe*UMAX/(NU/COARSEN));
  }

  {
    static double best = DBL_MAX;
    if (res_fe < best) {
      FILE *fp = fopen("best", "w");

      best = res_fe;
      fprintf(fp, "# %g\n", best);
      for (i=0; i<=DEGREE; ++i) {
	fprintf(fp, "%+g*x**%d", coeff[i], i);
      }
      fputs("\n", fp);
      fclose(fp);
    }
    for (i=0; i<=DEGREE; ++i) {
      printf("%g ", coeff[i]);
    }
    printf(" --> %g\n", res_fe);
  }

  xfree(x);
  return  res_fe;
}


int
main(int argc, char **argv)
{
  double *coeff, *eps;
  double res;
  const int l = COARSEN;
  const int nn = NU/l;
  int i;

  setpriority(PRIO_PROCESS, 0, PRIO_MAX);

  x_fe = xnew(double, nn);
  wrong_fe(NU/COARSEN, x_fe, COARSEN);

  {
    FILE *fp = fopen("path.dat", "w");
    int i;

    for (i=0; i<nn; ++i) {
      fprintf(fp, "%f %f\n", i*UMAX/nn, x_fe[i]);
    }
    fclose(fp);
  }

  coeff = xnew(double, DEGREE+1);
  if (try_read(argv[0], DEGREE, coeff) != 0) {
    for (i=0; i<=DEGREE; ++i)  coeff[i] = 0;
  }
  eps = xnew(double, DEGREE+1);
  for (i=0; i<=DEGREE; ++i)  eps[i] = 0.2;

  res = simplex(DEGREE+1, coeff, eps, 0.001, corrected, NULL);
  puts("--------------------");
  for (i=0; i<=DEGREE; ++i) {
    printf("%g ", coeff[i]);
  }
  printf(" --> %g\n", res);

  xfree(x_fe);
  return 0;
}
