/* simplex.c - the downhill simplex method of Nelder and Mead (parallel version)
 *
 * Copyright (C) 2010  Jochen Voss
 * Copyright (C) 2006  Jochen Voss, Andreas Voss
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include "burgers.h"


struct pv {
  double (*fn)(const double *, void *);
  double *x;
  double *data;
  double  val;
};

static void
evaluate(void *data)
{
  struct pv *point = data;
  point->val = point->fn(point->x, point->data);
}


static int
compare_points(const void *a, const void *b)
{
  const struct pv *pa = a;
  const struct pv *pb = b;

  if (pa->val < pb->val)
    return -1;
  if (pa->val > pb->val)
    return +1;
  return 0;
}


static void
move_point(int n, double p, const double *x, const double *y, double *z)
/* Move a point in direction of another point. x, y and z must be
 * vectors of length n. The function reads x and y and stores x +
 * p*(y-x) in z. For p=2 this calculates the reflection of x at y.  */
{
  int  i;
  for (i = 0; i < n; ++i)
    z[i] = x[i] + p*(y[i]-x[i]);
}

double
simplex(int n, double *x, const double *eps, double size_limit,
	double (*fn)(const double *x, void *data), void *data)
/* Multidimensional minimisation of the function 'fn' using the
 * downhill simplex algorithm of Nelder and Mead.
 *
 * 'fn' is the function to minimise.  The first argument of 'fn' is a
 * vector x which consists of 'n' doubles.  With the 'data' pointer
 * additional information may be forwarded to 'fn'.  The input values
 * of 'x' define one vertex of the starting simplex.  For the other
 * vertices, one of the input values of x is moved by the according
 * value of 'eps'.  'size_limit' defines the precision of the
 * search. The algorithm is terminated as soon as all edges of the
 * simplex are smaller than 'size_limit'.  On return the vector 'x'
 * contains the best found solution, that is, the parameters with the
 * smallest result of 'fn'.
 */
{
  struct factory *f;
  struct pv *points, tmp1, tmp2;
  double *mid;
  double  res;
  int  i, j, k;

  assert(n >= 2);
  f = new_factory(0);

  /* initialise the simplex */
  points = xnew(struct pv, n+1);
  for (j=0; j<=n; ++j) {
    points[j].fn = fn;
    points[j].x = xnew(double, n);
    for (i=0; i<n; ++i) {
      points[j].x[i] = x[i] + (i==j ? eps[i] : 0);
    }
    points[j].data = data;
    factory_job_start(f, evaluate, points+j);
  }
  tmp1.fn = fn;
  tmp1.x = xnew(double, n);
  tmp1.data = data;
  tmp2.fn = fn;
  tmp2.x = xnew(double, n);
  tmp2.data = data;

  mid = xnew(double, n);

  for (k = 0;; ++k) {
    for (i=0; i<=n; ++i)  factory_job_finish(f, points+i);
    qsort(points, n+1, sizeof(struct pv), compare_points);

    /* stopping criteria: by iterations ... */
    if (k > 10000) break;

    /* ... and by simplex size */
    {
      double  size = 0.0;
      for (i = 0; i < n; ++i) {
	double  min, max;

	min = max = points[0].x[i];
	for (j=1; j<=n; ++j) {
	  double  x = points[j].x[i];
	  if (x < min) {
	    min = x;
	  } else if (x > max) {
	    max = x;
	  }
	}
	if (max - min > size)
	  size = max - min;
	if (size > size_limit)
	  break;
      }
      if (size < size_limit)  break;
    }

    /* calculate the reflection point */
    for (i=0; i<n; ++i)  mid[i] = points[0].x[i];
    for (j=1; j<n; ++j) {
      for (i=0; i<n; ++i)  mid[i] += points[j].x[i];
    }
    for (i=0; i<n; ++i)  mid[i] /= n;

    /* try to reflect */
    move_point(n, -1, mid, points[n].x, tmp1.x); /* reflection */
    factory_job_start(f, evaluate, &tmp1);
    move_point(n, -2, mid, points[n].x, tmp2.x); /* extended reflection */
    factory_job_start(f, evaluate, &tmp2);
    factory_job_finish(f, &tmp1);
    factory_job_finish(f, &tmp2);

    if (compare_points(&tmp1, &tmp2) < 0) {
      if (compare_points(&tmp1, points+(n-1)) < 0) {
	/* replace the worst point with the reflection */
	memcpy(points[n].x, tmp1.x, n*sizeof(double));
	points[n].val = tmp1.val;
	continue;
      }
    } else {
      if (compare_points(&tmp2, points+(n-1)) < 0) {
	/* replace the worst point with the extended reflection */
	memcpy(points[n].x, tmp2.x, n*sizeof(double));
	points[n].val = tmp2.val;
	continue;
      }
    }

    /* try to contract */
    move_point(n, 0.5, mid, points[n].x, tmp1.x);
    evaluate(&tmp1);
    if (compare_points(&tmp1, points+n) < 0) {
      /* replace the worst point with the contraction */
      memcpy(points[n].x, tmp1.x, n*sizeof(double));
      points[n].val = tmp1.val;
      continue;
    }

    /* if everything else fails, use the multiple contraction */
    for (j = 1; j <= n; ++j) {
      /* move point j halfways towards the best point */
      move_point(n, 0.5, points[0].x, points[j].x, points[j].x);
      factory_job_start(f, evaluate, points+j);
    }
  }

  memcpy(x, points[0].x, n*sizeof(double));
  res = points[0].val;

  xfree(tmp1.x);
  xfree(tmp2.x);
  xfree(mid);
  for (j=0; j<=n; ++j) {
    xfree(points[j].x);
  }
  xfree(points);
  delete_factory(f);

  return res;
}
