/* burgers.c - main program to generate figure 7
 *
 * Copyright (C) 2010  Jochen Voss
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA.
 */

#include <stdio.h>
#include <math.h>
#include <float.h>
#include <assert.h>

#include "burgers.h"
#include "random4.h"


#ifndef METHOD
#define METHOD 1
#endif


static double
initial(double x)
{
  return -1 + 0.2*sin(x);
}

struct run_data {
  int n;
  double val;			/* constant c or parameter epsilon */
  double *x;
};

static void
reference_sln(void *data)
{
  struct run_data *rdata = data;
  int n = rdata->n;
  double c = rdata->val;
  double *x = rdata->x = xnew(double, n);

  double du = UMAX / n;
  unsigned long int  k, kk;
  int  i;

  assert(5*DT < du);	    /* plausibility check for CFL condition */

  /* initialisation */
  for (i=0; i<n; ++i)  x[i] = initial(i*du);

  /* main loop */
  kk = TMAX/DT+0.5;
  for (k=0; k<kk; ++k) {
    double  xl, xm, xr, x0;

    xm = x[n-1], x0 = xr = x[0];
    for (i=0; i<n; ++i) {
      xl = xm, xm = xr, xr = (i+1<n) ? x[i+1] : x0;
      double x2 = (xr+xm > 0) ? xm*xm : xr*xr;
      double x1 = (xm+xl > 0) ? xl*xl : xm*xm;
      x[i] += -(x2-x1)*(.5*DT/du); /* upwind scheme */
      x[i] += c * (DT/4);
      assert(! isnan(x[i]));
    }
  }
}

static void
FD_sln(void *data)
{
  struct run_data *rdata = data;
  int n = rdata->n;
  double eps = rdata->val;
  double *x = rdata->x = xnew(double, n);

  struct jvrand *rng;
  double *y;
  double du = UMAX / n;
  const double  gamma = THETA*DT*eps/(du*du);
  const double  delta = (1-THETA)*DT*eps/(du*du);
  const double  spde_sigma = sqrt(eps*DT/du);
  unsigned long int  k, kk;
  int  i;

  assert(5*DT < du);	    /* plausibility check for CFL condition */

  /* initialisation */
  rng = new_jvrand(SEED);
  for (i=0; i<n; ++i)  x[i] = initial(i*du);
  y = xnew(double, n);

  /* main loop */
  kk = TMAX/DT+0.5;
  for (k=0; k<kk; ++k) {
    double  xl, xm, xr;

    /* apply (I + nu (1-theta) DT L) */
    xmul(n, delta, 1-2*delta, delta, x, y);

    /* add drift and noise */
    xm = x[n-1], xr = x[0];
    for (i=0; i<n; ++i) {
      xl = xm, xm = xr, xr = (i+1<n) ? x[i+1] : x[0];
      y[i] += -xm*(xr-xm)*(DT/du); /* right hand derivative */
      y[i] += spde_sigma * jvrand_standard_normal(rng);
    }

    /* apply (I - nu theta DT L)^{-1} */
    solve(n, -gamma, 1+2*gamma, -gamma, y, x);
  }
  delete_jvrand(rng);

  xfree(y);
}

static double
dist(int n, const double *x, const double *y)
/* compute the L^2-distance between two solutions */
{
  double res;
  int i;

  if (isnan(x[0])) {
    res = DBL_MAX;
  } else {
    res = 0;
    for (i=0; i<n; i+=1) {
      double d = y[i] - x[i];
      res += d*d;
    }
    res = sqrt(res*UMAX/n);
  }
  return res;
}

int
main(int argc, char **argv)
{
  struct factory *f;
  int  k;

  f = new_factory(0);

  if (argc < 2) {
    fprintf(stderr, "warning: no deltas given, nothing to do\n");
  }

  for (k=1; k<argc; ++k) {
    char *tail;
    struct run_data  ref0, ref1;
    double log_delta, delta;
    double q = -1;
    int i, j, n;

    log_delta = strtod(argv[k], &tail);
    if (tail == argv[k] || *tail) {
      fprintf(stderr, "error: invalid value log10(delta)=%s, ignored\n",
	      argv[k]);
      continue;
    }
    delta = pow(10, log_delta);
    char out_name[32];
    snprintf(out_name, 32, "res%s.dat", argv[k]);
    FILE *out = fopen(out_name, "w");
    fprintf(stderr, "%s (delta=%g):\n", out_name, delta);

    n = UMAX/delta+.5;

    /* compute the viscosity solutions */
    ref0.n = n;
    ref0.val = 0.0;
    factory_job_start(f, reference_sln, &ref0);
    ref1.n = n;
    ref1.val = 1.0;
    factory_job_start(f, reference_sln, &ref1);

    /* compute the solutions with noise */
    int m = 22;
    struct run_data *epsdata = xnew(struct run_data, m);
    for (j=0; j<m; ++j) {
      epsdata[j].n = n;
      epsdata[j].val = pow(10, -.25*(j+3));
      factory_job_start(f, FD_sln, epsdata+j);
    }
    for (j=0; j<m; ++j) {
      double d0, d1;

      if (q < 0) {
	factory_job_finish(f, &ref0);
	factory_job_finish(f, &ref1);
	q = dist(n, ref0.x, ref1.x);

	if (k == argc-1) {
	  FILE *fp = fopen("reference.dat", "w");
	  for (i=0; i<n; ++i) {
	    fprintf(fp, "%f %f %f\n", i*UMAX/n, ref0.x[i], ref1.x[i]);
	  }
	  fprintf(fp, "%f %f %f\n", UMAX, ref0.x[0], ref1.x[0]);
	  fclose(fp);
	}
      }

      /* output the results as they come in */
      factory_job_finish(f, epsdata+j);
      d0 = dist(n, epsdata[j].x, ref0.x);
      d1 = dist(n, epsdata[j].x, ref1.x);
      fprintf(out, "%g %g %g\n", epsdata[j].val, d0/q, d1/q);
      fprintf(stderr, "%g %g %g\n", epsdata[j].val, d0/q, d1/q);
      fflush(NULL);
      if (k == argc-1 && j == m-1) {
	FILE *fp = fopen("path.dat", "w");
	for (i=0; i<n; ++i) {
  	fprintf(fp, "%f %f\n", i*UMAX/n, epsdata[j].x[i]);
	}
	fprintf(fp, "%f %f\n", UMAX, epsdata[j].x[0]);
	fclose(fp);
      }

      xfree(epsdata[j].x);
    }

    fputc('\n', stderr);

    xfree(epsdata);
    xfree(ref1.x);
    xfree(ref0.x);
    fclose(out);
  }

  delete_factory(f);
  return 0;
}
